-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

-- === GAME ENGINE

local M = {}

-- memory variables
local spawnSeed
local wearSeed
local lastConfig
local wearInfo
local isLastGenClone = false
local savedSettings = {}

-- cleanup functions
local function configBool(val,def)
	if type(val) == "boolean" then
		return val
	else
		return def
	end
end

-- main function
local function spawnBarnfind(genConfig, cloneVehicle) 
	log('I', 'Barnfind_State', 'Generating barnfind...')
	local carInfo = {}
	wearInfo = nil
	local success,err = pcall(function()
		-- remember configs and setup rng seed
		local globalSeed = os.time() + os.clock() * 1000
		spawnSeed = genConfig.Seed or globalSeed
		wearSeed = genConfig.WearSeed or globalSeed
		
		-- setup wear seed stuff
		math.randomseed(wearSeed)
		local conf_mileage = genConfig.Mileage or math.random(0,650000)
		local conf_condition = genConfig.Condition or math.random()
		math.randomseed(spawnSeed)
		
		lastConfig = genConfig
		isLastGenClone = cloneVehicle
		-- setup configs and other variables
		local conf_maxYear = genConfig.MaxYear or -1
		local conf_showInfo = configBool(genConfig.ShowInfo,true)
		local conf_showDistance = configBool(genConfig.ShowDist,true)
		local conf_showStateReport = configBool(genConfig.ShowState,true)
		local conf_wearVariation = genConfig.WearVar or .35
		local conf_condOverride = genConfig.Override or {}
		local conf_balanceWear = configBool(genConfig.Balance,true)
		local conf_usePopulation = configBool(genConfig.UsePopulation,true)
		local conf_WearRandomSeed = genConfig.WearSeed or (genConfig.Seed or globalSeed)
		local conf_SpawnRandomSeed = genConfig.Seed or globalSeed
		local conf_chancePark = genConfig.ParkChance or .15
		local conf_moddedVehicles = configBool(genConfig.ModCars,false)
		local conf_spawnAtCamera = configBool(genConfig.SpawnHere,false)
		local conf_showSeed = configBool(genConfig.ShowSeed,false)
		local conf_deform = configBool(genConfig.Deformation,false)
		
		local carPrefixA, carPrefixB
		if cloneVehicle then
			-- clone the player vehicle
			local veh = be:getPlayerVehicle(0)
			if not veh then
				guihooks.message("ERROR: There are no vehicles in the level!", 8, "barnfindGenerator.error.noVehiclesFound")
				log('E', 'Barnfind_Error', "Execution Failed: No vehicles selected.")
				error("exeFailed")
				return
			end
	
			-- we get the current vehicles parameters and feed it into the spawning function
			local metallicPaintData = veh:getMetallicPaintData()
			local options = {
				model  = veh.JBeam,
				config = veh.partConfig,
				paint  = createVehiclePaint(veh.color, metallicPaintData[1]),
				paint2 = createVehiclePaint(veh.colorPalette0, metallicPaintData[2]),
				paint3 = createVehiclePaint(veh.colorPalette1, metallicPaintData[3])
			}

			carPrefixA = veh.JBeam
			carPrefixB = options
		else
		
			local allCars = core_vehicles.getVehicleList().vehicles 
			local carModel, carConfig, paintColor, paintName
			
			-- get combined car population of all vehicles OR every configs
			local maxPop = 0 
			local possibleCars = {}
			for a,b in ipairs(allCars) do 
				if (conf_maxYear == -1 or (b.model.Years and b.model.Years.min <= conf_maxYear)) and (b.model.Type == "Car" or b.model.Type == "Truck") then 
					for c,d in pairs(b.configs) do 
						if not (d.Type and string.find(d.Type,"Prop")) and (conf_moddedVehicles or d.Source == "BeamNG - Official") then 
							if conf_usePopulation then
								if d.Population then
									possibleCars[c] = {b,d}
									maxPop = maxPop + d.Population 
								end
							else
								possibleCars[c] = {b,d}
							end
						end 
					end 
				end 
			end 
			
			if tableSize(possibleCars) == 0 then
				guihooks.message("ERROR: Couldn't find a vehicle with the selected configs.", 8, "barnfindGenerator.error.noVehiclesFitsConfigs")
				log('E', 'Barnfind_Error', "Execution Failed: Can't find a vehicle with the selected configs!")
				error("exeFailed")
				return
			end

			-- roll a random car to select depending on how populated it is
			local carName, carData
			if conf_usePopulation then
				local numLeft = math.random(1, maxPop)
				while not carData do
					for a,b in pairs(possibleCars) do
						numLeft = numLeft - b[2].Population
						if numLeft <= 0 then
							carName = a
							carData = b
							break
						end
					end
				end
			else
				local carRng = math.random(1, tableSize(possibleCars))  
				local carKeys = tableKeys(possibleCars)
				carName = carKeys[carRng]
				carData = possibleCars[carName]
			end
			
			local modelData = carData[1].model
			local variantData = carData[2]
									
			local years = variantData.Years or modelData.Years or nil
			-- randomize their paint with available factory colors
			local allPaints = modelData.paints 
			local paintRng = math.random(1, tableSize(allPaints))  
			
			local allPaintsKeys = tableKeys(allPaints)
			paintName = allPaintsKeys[paintRng]
			paintColor = allPaints[paintName]
									
			-- gather vehicle info
			carInfo = {Vehicle = (modelData.Brand or "").." "..modelData.Name,
			Variant = variantData.Configuration or "N/A",
			Year = (years and math.random(years.min,years.max) or "N/A"),
			Color = paintName or "N/A",
			Population = variantData.Population or "N/A",
			Value = variantData.Value and "$ "..tostring(variantData.Value) or "N/A"}
			
			carPrefixA = modelData.key
			carPrefixB = {config = carName, paint = paintColor}
		end
		
		local park, road
		if not conf_spawnAtCamera then
			-- find a random parking spot to spawn the vehicle
			if math.random() < conf_chancePark then
				local allParks = gameplay_parking.getParkingSpots() -- this function makes sure parking sites are loaded before being returned
				local parkingSpots = gameplay_parking.findParkingSpots(nil, nil, 10000000000) 
				local parkingSpotNames = tableKeys(parkingSpots) 

				local parkingSpotName = parkingSpotNames[math.random(tableSize(parkingSpotNames))] 
				park = parkingSpots[parkingSpotName]
				
				if not park then
					log('W', 'Barnfind_Warning', "This map doesn't have any parking spots, spawning vehicle on a road instead...")
				end
			end
			
			if park then
				local approxDistance = math.floor(math.sqrt(park.squaredDistance))
				if conf_showDistance then
					carInfo['Distance'] = tostring(approxDistance).." m"
				end
			else
				-- if not, find a random road to spawn the vehicle instead
				local allNodes = map.getMap().nodes
				local pathNodes = tableSize(allNodes)
				
				if pathNodes > 0 then
					local selRoad = math.random(1,pathNodes)
					local allNodesKeys = tableKeys(allNodes)
					
					local roadName = allNodesKeys[selRoad]
					road = allNodes[roadName]
					
					local approxDistance = math.floor(math.sqrt(road.pos:squaredDistance(core_camera.getPosition())))
					if conf_showDistance then
						carInfo['Distance'] = tostring(approxDistance).." m"
					end
				else
				
					guihooks.message("ERROR: No available locations could be found to spawn the vehicle. You might need to use a different map.", 8, "barnfindGenerator.error.noSpawnLocation")
					log('E', 'Barnfind_Error', "Execution Failed: This map does't contain any roads to spawn a vehicle.")
					error("exeFailed")
					return
				end
			end
		end

		-- spawn the vehicle at the selected parking spot & setup the rest
		local car = core_vehicles.spawnNewVehicle(carPrefixA, carPrefixB) 
		be:enterNextVehicle(0, 1) 
		
		if park then
			gameplay_parking.moveToParkingSpot(car:getId(), park.ps, true)
		else
			local vec = road and road.pos + vec3(0,0,.5) or core_camera.getPosition()
			local qua = road and quatFromEuler(0, 0, math.random() * 360) or quat(0, 0, 0, 1)
			car:setPosRot(vec.x, vec.y, vec.z, qua.x, qua.y, qua.z, qua.w)
			
			car:queueLuaCommand("electrics.setIgnitionLevel(0)")
		end
		
		spawn.safeTeleport(car, car:getPosition(), car:getRotation())
		
		-- send the vehicle and condition info to vehicle lua
		local stringTable = "{"
		local tabSize = tableSize(conf_condOverride)
		local ind = 1
		for a,b in pairs(conf_condOverride) do
			stringTable = stringTable..a.." = "..b
			ind = ind + 1
			if ind <= tabSize then
				stringTable = stringTable..", "
			end
		end
		
		local strWearSeed = tostring(wearSeed)
		local strConfMile = tostring(conf_mileage * 1000)
		local strConfCond = tostring(conf_condition)
		local strConfWearVar = tostring(conf_wearVariation)
		local strShowRep = tostring(conf_showStateReport)
		local strConfBal = tostring(conf_balanceWear)
		local strDeform = tostring(conf_deform)
		
		car:queueLuaCommand("(function() local condOverride = "..stringTable.."} return extensions.barnfindGenerator.setupBarnfind("..strWearSeed..","..strConfMile..","..strConfCond..","..strConfWearVar..","..strShowRep..",condOverride,"..strConfBal..", true, "..strDeform..") end) ()") 

		-- show the vehicle info on the console
		carInfo["Mileage"] = tostring(math.floor(conf_mileage)).." km"
		carInfo["Condition"] = tostring(math.floor(conf_condition * 100)).." %"
		if conf_showSeed then
			carInfo["Seed"] = tostring(math.floor(conf_SpawnRandomSeed))
		end
		
		if conf_showInfo then
			for a,b in pairs(carInfo) do
				log('O', 'Barnfind_Info', tostring(a)..": "..tostring(b))
			end
		end
	end)
	
	-- protected call result
	if success then
		guihooks.message("Successfully generated a vehicle!", 4, "barnfindGenerator.success.spawnGenerator")
		log('I', 'Barnfind_State', 'Execution Complete!')

		return carInfo
	elseif not string.find(err,"exeFailed") then
		guihooks.message("ERROR: An unexpected lua error occured, check console for more info.", 8, "barnfindGenerator.error.luaError")
		log('E', 'Barnfind_Error', 'Execution Failed due to unexpected Lua Error: '..err)
	end
end

-- applying wear to current vehicle
local function applyWear(genConfig)
	local conf_WearRandomSeed = (genConfig.WearSeed or genConfig.Seed) or os.time() + os.clock() * 1000
	math.randomseed(conf_WearRandomSeed)
	
	local conf_condition = genConfig.Condition or math.random()
	local conf_mileage = genConfig.Mileage or math.random(0,650000)
	local conf_showStateReport = configBool(genConfig.ShowState,true)
	local conf_wearVariation = genConfig.WearVar or .35
	local conf_condOverride = genConfig.Override or {}
	local conf_balanceWear = configBool(genConfig.Balance,true)
	local conf_showSeed = configBool(genConfig.ShowSeed,false)
	local conf_deform = configBool(genConfig.Deformation,false)
	local strDeform = tostring(conf_deform)
	local strWearSeed = tostring(conf_WearRandomSeed)
	local strConfMile = tostring(conf_mileage * 1000)
	local strConfCond = tostring(conf_condition)
	local strConfWearVar = tostring(conf_wearVariation)
	local strShowRep = tostring(conf_showStateReport)
	local strConfBal = tostring(conf_balanceWear)
	
	local stringTable = "{"
	local tabSize = tableSize(conf_condOverride)
	local ind = 1
	for a,b in pairs(conf_condOverride) do
		stringTable = stringTable..a.." = "..b
		ind = ind + 1
		if ind <= tabSize then
			stringTable = stringTable..", "
		end
	end
	
	local car = be:getPlayerVehicle(0)
	if car then
		car:queueLuaCommand("(function() local condOverride = "..stringTable.."} return extensions.barnfindGenerator.setupBarnfind("..strWearSeed..","..strConfMile..","..strConfCond..","..strConfWearVar..","..strShowRep..",condOverride,"..strConfBal..", true, "..strDeform..") end) ()") 
		guihooks.message("Successfully applied vehicle condition!", 4, "barnfindGenerator.success.applyConditions")
		
		local returninfo = {Condition = tostring(math.floor(conf_condition * 100)).." %", Mileage = tostring(math.floor(conf_mileage)).." km"}
		if conf_showSeed then
			returninfo["Seed"] = tostring(math.floor(conf_WearRandomSeed))
		end
		return returninfo
	else
		guihooks.message("ERROR: There are no vehicles in the level!", 8, "barnfindGenerator.error.noVehiclesFound")
		log('E', 'Barnfind_Error', "Execution Failed: No vehicles selected.")
	end
end

-- respawn previous generated vehicle
local function respawnPrev()
	if lastConfig then
		lastConfig["Seed"] = spawnSeed
		lastConfig["WearSeed"] = wearSeed
		
		local info = spawnBarnfind(lastConfig, isLastGenClone)
		if info then
			guihooks.message("Successfully regenerated previous vehicle!", 4, "barnfindGenerator.success.spawnGenerator")
		end
		return info
	else
		guihooks.message("ERROR: No vehicles were previously generated!", 8, "barnfindGenerator.error.noPreviousGenerations")
		log('E', 'Barnfind_Error', "Execution Failed: No vehicle previously generated.")
	end
end

-- setting up paint colors
local function setupPaint(vehId, paint_cond, allData)
	local success,err = pcall(function()
		local daCar = be:getObjectByID(vehId)
		for a,b in ipairs(allData) do
			local paintData = daCar:getMeshColor(b)
			local col1 = paintData[1]:toTable()
			local col2 = paintData[2]:toTable()
			local col3 = paintData[3]:toTable()
			local fullTable = {col1[1],col1[2],col1[3], col2[1],col2[2],col2[3], col3[1],col3[2],col3[3]}
			
			local colRNG = linearScale(paint_cond, 0, .75, .25, 0, 0) < math.random()
			for c,d in ipairs(fullTable) do
				local curCol = colRNG and d or math.random(0,255)		
				local diff = curCol - 90
				fullTable[c] = curCol - (diff * math.max(linearScale(paint_cond, .5, 0, 0, 1.15), 0))
			end
			
			daCar:setMeshColor(b, ColorI(fullTable[1], fullTable[2], fullTable[3], 255), ColorI(fullTable[4], fullTable[5], fullTable[6], 255), ColorI(fullTable[7], fullTable[8], fullTable[9], 255))
			daCar:setMeshPaintData(b, ColorI(fullTable[1], fullTable[2], fullTable[3], 255), ColorI(fullTable[4], fullTable[5], fullTable[6], 255), ColorI(fullTable[7], fullTable[8], fullTable[9], 255))
		end
	end)
	
	if not success then
		uihooks.message("ERROR: Failed to apply vehicle paint.", 8, "barnfindGenerator.error.vehiclePaintFailed")
		log('E', 'Barnfind_Error', "Execution Failed: Failed to apply paint.")
	end
end

-- functions for handling wait times for gathering wear info.
local function vehicleInfoCallback(info)
	wearInfo = info
end

local function getWearInfo()
	return wearInfo
end

-- ui settings
local function loadSettings()
	return savedSettings
end

local function saveSettings(settings)
	savedSettings = settings
end

M.respawnPrev = respawnPrev
M.spawnBarnfind = spawnBarnfind
M.applyWear = applyWear

M.loadSettings = loadSettings
M.saveSettings = saveSettings

M.setupPaint = setupPaint
M.getWearInfo = getWearInfo
M.vehicleInfoCallback = vehicleInfoCallback

return M