angular.module('beamng.apps')
  .directive('usedCarGenerator', [function () {
    return {
      templateUrl:
        '/ui/modules/apps/usedCarGenerator/app.html',
      replace: true,
      restrict: 'EA',
      scope: true,
      controller: ['$log', '$scope', function ($log, $scope) {
        // Main data stuff
        var getDefaultSettings = function () {
          return {
            setting_condition: null,
            setting_mileage: null,
            setting_parkchance: 0.15,
            setting_usepopulation: true,
            setting_moddedcars: false,
            setting_maxyear: null,
            setting_seed: null,
            
            setting_wearseed: null,
            setting_balance: true,
            setting_wearvariation: 0.35,
            setting_deformation: false,
            setting_showinfo: true,
            setting_showvehiclestate: true,
            setting_showdistance: true,
            setting_showseed: false,

            setting_override: {
              Paint: null, Panels: null, Fixtures: null, Brakes: null, Suspension: null, Tires: null,
              Radiator: null, Coolant: null, HeadGasket: null, PistonRings: null, ConnectingRods: null, Oilpan: null,
              FuelLevel: null, Exhaust: null, Crankshaft: null, SparkPlugs: null, FuelPump: null, IdleController: null,
              ElectricMotors: null, Battery: null, Supercharger: null, TurboTurbine: null, TurboCompressor: null,
              ClutchPlate: null, ClutchDisc: null, ClutchSprings: null, Gearbox: null
            }
          }
        }

        var updateConfigs = function() {
          return {
            Condition: $scope.settings.setting_condition,
            Mileage: $scope.settings.setting_mileage,
            ParkChance: $scope.settings.setting_parkchance,
            UsePopulation: $scope.settings.setting_usepopulation,
            ModCars: $scope.settings.setting_moddedcars,
            MaxYear: $scope.settings.setting_maxyear,
            Seed: $scope.settings.setting_seed,
            WearSeed: $scope.settings.setting_wearseed,
            Balance: $scope.settings.setting_balance,
            WearVar: $scope.settings.setting_wearvariation,
            Deformation: $scope.settings.setting_deformation,
            ShowInfo: $scope.settings.setting_showinfo,
            ShowState: $scope.settings.setting_showvehiclestate,
            ShowDist: $scope.settings.setting_showdistance,
            ShowSeed: $scope.settings.setting_showseed,
            Override: $scope.settings.setting_override
          }
        }

        $scope.settings = getDefaultSettings()

        $scope.displayOverrideInfo = [];
        $scope.displayInfo = [];
        $scope.minimizeTab = false;

        var displayOverride = false;
        var configs = {};
        var attempts = 0;

        // Show the override menu
        $scope.toggleOverrides = () => {
          displayOverride = !displayOverride
          let overrideInfo = []

          if (displayOverride) {
            let allOverrides = getDefaultSettings().setting_override
            for (const key in allOverrides) {
              if ($scope.settings.setting_override[key] == undefined) {
                $scope.settings.setting_override[key] = null;
              }
            }
            
            let infoIndex = 0
            let index = 0
            for (const key in $scope.settings.setting_override) {
              if (index % 2 == 0) {
                overrideInfo[infoIndex] = {k1: key, k2: null}
              } else {
                overrideInfo[infoIndex] = {k1: overrideInfo[infoIndex].k1, k2: key}
                infoIndex++
              }
              index++
            }
          }

          $scope.$evalAsync(function () {
            $scope.displayOverrideInfo = overrideInfo
          })
        }

        // The rest of the UI button functions
        $scope.spawnBarnfind = (spawnHere, cloneVehicle) => {
          configs = updateConfigs()
          let str = convertStringData(configs)
          sendEngineLua(`(function() return barnfindGenerator.spawnBarnfind({${str}, SpawnHere = ${spawnHere}}, ${cloneVehicle}) end) ()`)
        }

        $scope.applyCurrentVehicle = () => {
          configs = updateConfigs()
          let str = convertStringData(configs)
          sendEngineLua(`(function() return extensions.barnfindGenerator.applyWear({${str}}) end) ()`)
        }

        $scope.regenPrevious = () => {
          sendEngineLua(`(function() return extensions.barnfindGenerator.respawnPrev() end) ()`)
        }

        $scope.reapplyConditions = () => {
          bngApi.engineLua(`local veh = be:getPlayerVehicle(0)
            if veh then
             veh:queueLuaCommand("extensions.barnfindGenerator.resetBarnfind()")
            else
              guihooks.message("ERROR: There are no vehicles in the level!", 8, "barnfindGenerator.error.noVehicleFound")
            end`
          )
        }

        $scope.resetSettings = () => {
          $scope.settings = getDefaultSettings()
        }

        $scope.minimize = () => {
          $scope.$evalAsync(function () {
            $scope.minimizeTab = !$scope.minimizeTab
          })
        }

        // Convert settings data to Lua string
        var convertStringData = function(data) {
          let str = ""
          let first = true
          for (const key in data) {
            if (!first) { 
              str += `,` 
            }
            first = false

            let writeVal = data[key];
            if (typeof data[key] == "object" && data[key] != null) {
              writeVal = "{" + convertStringData(data[key]) + "}"
            }
            str += `${key} = ${writeVal != null ? writeVal : 'nil'}`
          }
          return str
        }

        // Main function to execute a barnfind generation
        var sendEngineLua = function (luaScript) {
          makeDataDisplayable([{ name1: "Loading...", val1: "", name2: "Loading...", val2: "" }])
          configs = updateConfigs()

          setTimeout(() => bngApi.engineLua(luaScript, function (data) {
            if (data == null) { 
              makeDataDisplayable([{ name1: "Loading Failed.", val1: "", name2: "Loading Failed.", val2: "" }])
              return 
            }
            let keyInfo = Object.keys(data)
            let valInfo = Object.values(data)

            attempts = 0;
            setTimeout(function () {
              queueWearInfo(keyInfo, valInfo)
            }, 100);
          }), 50)
        }

        // Wait for the script to give us back the conditions
        var queueWearInfo = function (keyInfo, valInfo) {
          setTimeout(() => bngApi.engineLua(`(function() return barnfindGenerator.getWearInfo() end) ()`, function (data) {
            if (data == null) {
              if (attempts < 50) {
                attempts++
                queueWearInfo()
              } else {
                makeDataDisplayable([{ name1: "Loading Failed.", val1: "", name2: "Loading Failed.", val2: "" }])
              }
              return
            }

            let keyWear = Object.keys(data)
            let valWear = Object.values(data)
            makeDataDisplayable(updateDisplayData(keyInfo, valInfo, keyWear, valWear))
          }), 100)
        }

        // Update the UI Info text
        var updateDisplayData = function (ki, vi, kw, vw) {
          let fullData = []

          for (let i = 0; i < Math.max(ki.length, kw.length); i++) {
            let sI = configs.ShowInfo
            let sW = $scope.settings.setting_showvehiclestate
            let realKw = ""
            if (kw[i]) {
              let str = kw[i].replace("__", " - ")
              realKw = str.replace("_", " ")
            }

            fullData[i] = { name1: (sI && ki[i] ? ki[i] : null), val1: (sI && vi[i] ? vi[i] : null), name2: (sW && kw[i] ? realKw : null), val2: (sW && vw[i] ? vw[i] : null) }
          }
          return fullData
        }  

        var makeDataDisplayable = function (display) {
          $scope.$evalAsync(function () {
            $scope.displayInfo = display
          })
        }

        // Load settings when the UI is first loaded
        setTimeout(() => bngApi.engineLua(`(function() local a,b = pcall(function() return extensions.barnfindGenerator.loadSettings() end) return a and b or nil end) ()`, function (data) {
          if (data == null) {
            return
          }
          let keySets = Object.keys(data)
          let valSets = Object.values(data)

          $scope.$evalAsync(function () {
            for (let i = 0; i < keySets.length; i++) {
              $scope.settings[keySets[i]] = valSets[i]
            }
          })
        }), 50)

        // Save settings when the UI is unloaded
        $scope.$on('$destroy', function () {
          let str = convertStringData($scope.settings)
          bngApi.engineLua(`(function() return pcall(function() return extensions.barnfindGenerator.saveSettings({${str}}) end) end) ()`)
        })
      }
      ]
    }
  }]);